/* (c)2013 ALTIMESH & HighPerformanceComputing Magazine France - All rights reserved */

#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <stdio.h>
#include <sys/time.h>

struct timer
{
        struct timeval _start ;
        struct timeval _stop ;

        void start ()
        {
		::gettimeofday (&_start, 0) ;
        }
        void stop ()
        {
		::gettimeofday (&_stop, 0) ;
        }
        double duration ()
        {
		return (((double)_stop.tv_usec - (double)_start.tv_usec) * 1.0e-6) +
		(double)_stop.tv_sec - (double)_start.tv_sec ;
        }
} ;


__global__ void initkernel(int count, float * a, int numidx, int* index, float* b, float* c, float* d)
{
    for (int i = threadIdx.x + blockDim.x * blockIdx.x ; i < count ; i += blockDim.x * gridDim.x)
    {
		for (int k = 0 ; k < numidx ; ++k)
		{
            a [k * count + i] = 1.0f / (0.5f + (float)(i+k)) ;
			if (i == 0)
				index [k] = ((k * k) % numidx) * count ;
		}
        b [i] = 0.0f ;
        c [i] = 0.0f ;
        d [i] = 0.0f ;
    }
}

__global__ void latencykernel(int count, const float * a, int numidx, const int* index, float * b)
{
	for (int i = threadIdx.x + blockDim.x * blockIdx.x ; i < count ; i += blockDim.x * gridDim.x)
	{
		float tmp = 0.0 ;
		for (int k = 0 ; k < numidx ; ++k)
		{
			tmp += a [i + index [k]] ;
		}
		b [i] = tmp ;
	}
}

__global__ void latencykernelswap(int count, const float * a, int numidx, const int* index, float * b)
{
	for (int i = threadIdx.x + blockDim.x * blockIdx.x ; i < count ; i += blockDim.x * gridDim.x)
	{
		b [i] = 0.0f ;
	}
		
	for (int k = 0 ; k < numidx ; ++k)
	{
		for (int i = threadIdx.x + blockDim.x * blockIdx.x ; i < count ; i += blockDim.x * gridDim.x)
		{
			b [i] += a [i + index [k]] ;
		}
	}
}

__device__ float4 __ldg(const float4* a) ;
// 64 * (arraysize * 4) flops minimum for this kernel
__global__ void latencykernelfloat4(int count, const float * __restrict a, int numidx, const int* index, float* b)
{
	for (int i = 8 * (threadIdx.x + blockDim.x * blockIdx.x) ; i < count ; i += blockDim.x * gridDim.x * 8)
	{
		float4 leftP, rightP ;
		float4 left, right ;
		leftP.x = 0.0f ; leftP.y = 0.0f ; leftP.z = 0.0f ; leftP.w = 0.0f ;
		rightP.x = 0.0f ; rightP.y = 0.0f ; rightP.z = 0.0f ; rightP.w = 0.0f ;
		for (int k = 0 ; k < numidx ; ++k)
		{
			left = __ldg ((float4*)(&(a [i + index [k]]))) ;
			right = __ldg ((float4*)(&(a [i + 4 + index [k]]))) ;
			leftP.x += left.x ; leftP.y += left.y ; leftP.z += left.z ; leftP.w += left.w ; 
			rightP.x += right.x ; rightP.y += right.y ; rightP.z += right.z ; rightP.w += right.w ; 
		}
		*((float4*)(&b[i])) = leftP ;
		*((float4*)(&b[i+4])) = rightP ;
	}
}

void cuda_check_func (cudaError_t er, int line, const char* file)
{
       if (er != cudaSuccess) ::fprintf (stderr, "[%s:%d] -- Cuda error : %d %s\n", file, line, er, ::cudaGetErrorString (er)) ;
}

#define CUDA_CHECK(a) cuda_check_func(a,__LINE__,__FILE__)

int main()
{
	CUDA_CHECK(::cudaSetDevice (0)) ;

       cudaDeviceProp props ;
       CUDA_CHECK (::cudaGetDeviceProperties (&props, 0)) ;

	int nthreads = 1024 ;
	int nblocks = 8 * props.multiProcessorCount ;

	int numindex = 60 ;
	int* indices ;

       int arraySize = nblocks * 1024 ;
       float *da, *db, *dc, *dd ;
       CUDA_CHECK (::cudaMalloc (&da, arraySize * numindex * sizeof(float))) ;
       CUDA_CHECK (::cudaMalloc (&db, arraySize * sizeof(float))) ;
       CUDA_CHECK (::cudaMalloc (&dc, arraySize * sizeof(float))) ;
       CUDA_CHECK (::cudaMalloc (&dd, arraySize * sizeof(float))) ;
	   CUDA_CHECK (::cudaMalloc (&dd, arraySize * sizeof(float))) ;
	   CUDA_CHECK (::cudaMalloc (&indices, numindex * sizeof(int))) ;

       initkernel <<< nblocks, nthreads >>> (arraySize, da, numindex, indices, db, dc, dd) ;
	   CUDA_CHECK (::cudaDeviceSynchronize());

       timer plain, f4, v4 ;

	plain.start () ;

       latencykernel <<< nblocks, nthreads >>> (arraySize, da, numindex, indices, db) ;
       CUDA_CHECK (::cudaDeviceSynchronize());

	plain.stop () ;
	v4.start () ;

       latencykernelswap <<< nblocks, nthreads >>> (arraySize, da, numindex, indices, dc) ;
       CUDA_CHECK (::cudaDeviceSynchronize());

	v4.stop () ;
	f4.start () ;

       latencykernelfloat4 <<< nblocks, nthreads >>> (arraySize, da, numindex, indices, dd) ;
       CUDA_CHECK (::cudaDeviceSynchronize());

	f4.stop () ;

       float* b = new float[arraySize] ;
       float* c = new float[arraySize] ;
       float* d = new float[arraySize] ;

       CUDA_CHECK (::cudaMemcpy (b, db, sizeof (float) * arraySize, cudaMemcpyDeviceToHost)) ;
       CUDA_CHECK (::cudaMemcpy (c, dc, sizeof (float) * arraySize, cudaMemcpyDeviceToHost)) ;
       CUDA_CHECK (::cudaMemcpy (d, dd, sizeof (float) * arraySize, cudaMemcpyDeviceToHost)) ;

       ::fprintf (stdout, "naive, swap, float4\n") ;
       for (int k = 0 ; k < 100 ; ++k)
       {
             ::fprintf (stdout, "%f, %f, %f\n", b[k], c[k], d[k]) ;
       }

       double flopCount = 13.0 * 8.0 * (double)arraySize ;

       ::fprintf (stderr, "naive operation took %lf s\n", plain.duration ()) ;
       ::fprintf (stderr, "swap operation took %lf s\n", v4.duration ()) ;
       ::fprintf (stderr, "float4 operation took %lf s\n", f4.duration ()) ;

       ::fprintf (stderr, "naive elements/s >= %lf \n", (double) arraySize / plain.duration()) ;
       ::fprintf (stderr, "swap elements/s >= %lf \n", (double) arraySize / v4.duration()) ;
       ::fprintf (stderr, "float4 elements/s >= %lf \n", (double) arraySize / f4.duration ()) ;

	   double memoryread = arraySize * numindex * sizeof(float) / (1000.0*1000.0*1000.0) ;

       ::fprintf (stderr, "naive approx. read BW >= %lf GB/s \n", (double) memoryread / plain.duration()) ;
       ::fprintf (stderr, "swap approx. read BW >= %lf GB/s \n", (double) memoryread / v4.duration()) ;
       ::fprintf (stderr, "float4 approx. read BW >= %lf GB/s \n", (double) memoryread / f4.duration ()) ;

		::fprintf (stdout, "problem size, naive(s), swap(s), float4(s), naive(GB/s), swap(GB/s), float4(GB/s)\n") ;
	   for (int k = 32 ; k <= 1024 ; k += 32)
	   {
		   memoryread = ((arraySize*k)/1024) * numindex * sizeof(float) / (1000.0*1000.0*1000.0) ;

			plain.start () ;

			   latencykernel <<< nblocks, ((nthreads*k)/1024) >>> (((arraySize*k)/1024), da, numindex, indices, db) ;
			   CUDA_CHECK (::cudaDeviceSynchronize());

			plain.stop () ;
			v4.start () ;

			   latencykernelswap <<< nblocks, ((nthreads*k)/1024)  >>> (((arraySize*k)/1024), da, numindex, indices, dc) ;
			   CUDA_CHECK (::cudaDeviceSynchronize());

			v4.stop () ;
			f4.start () ;

			   latencykernelfloat4 <<< nblocks, ((nthreads*k)/1024)  >>> (((arraySize*k)/1024), da, numindex, indices, dd) ;
			   CUDA_CHECK (::cudaDeviceSynchronize());

			f4.stop () ;
			
			::fprintf (stdout, "%d, %lf, %lf, %lf, %lf, %lf, %lf\n", ((arraySize*k)/1024) * nblocks, plain.duration (), v4.duration (), f4.duration (),
						(double) memoryread / plain.duration(), (double) memoryread / v4.duration(), (double) memoryread / f4.duration ()) ;
	   }

    return 0;
}


