/* (c)2013 ALTIMESH & HighPerformanceComputing Magazine France - All rights reserved */

#include <cstdio>
#include <cstdlib>
#include <immintrin.h>
#include <sys/time.h>

#include <omp.h>

__attribute__((target(mic))) void addition_naive(int size, const float* __restrict a, float* __restrict c)
{
	#ifdef __MIC__
	#pragma omp parallel 
	{
		float accum = 0.0f ;
		int startK = omp_get_thread_num() * (size/16) / omp_get_num_threads() ;
		int endK = (1 + omp_get_thread_num()) * (size/16) / omp_get_num_threads() ;
		#pragma vector aligned
		for (int k = 16*startK ; k < 16*endK ; ++k)
		{
			accum += a [k] ;
		}
		c [omp_get_thread_num()] = accum ;
	}
	#else
	#endif
}

__attribute__((target(mic))) void addition_gather(int size, const float* __restrict a, const float* __restrict b, float* __restrict c)
{
	#ifdef __MIC__
	__m512i baseindex ;
	baseindex = _mm512_set_epi32(0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15);
	#pragma omp parallel 
	{
		int startK = omp_get_thread_num() * (size/16) / omp_get_num_threads() ;
		int endK = (1 + omp_get_thread_num()) * (size/16) / omp_get_num_threads() ;
		__m512i index ;
		__m512 accum = _mm512_set1_ps(0.0f) ;
		for (int k = startK ; k < endK ; ++k)
		{
			index = _mm512_set1_epi32(16*k);
			index = _mm512_add_epi32 (index, baseindex) ;
			// c [k] = a [k] + b [k] ;
			
			__m512 at = _mm512_i32gather_ps (index, a, sizeof (float)) ;
			__m512 bt = _mm512_i32gather_ps (index, b, sizeof (float)) ;
			accum = _mm512_add_ps (accum, at) ;
			accum = _mm512_add_ps (accum, bt) ;
		}
		index = _mm512_set1_epi32(16 * omp_get_thread_num());
		_mm512_i32scatter_ps (c, index, accum, sizeof (float)) ;
	}
	#else
	#endif
}

__attribute__((target(mic))) void addition_load(int size, const float* __restrict a, const float* __restrict b, float* __restrict c)
{
	#ifdef __MIC__
	__m512i baseindex ;
	baseindex = _mm512_set_epi32(0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15);
	#pragma omp parallel 
	{
		int startK = omp_get_thread_num() * (size/16) / omp_get_num_threads() ;
		int endK = (1 + omp_get_thread_num()) * (size/16) / omp_get_num_threads() ;
		__m512i index ;
		__m512 accum = _mm512_set1_ps(0.0f) ;
		for (int k = startK ; k < endK ; ++k)
		{
			index = _mm512_set1_epi32(16*k);
			index = _mm512_add_epi32 (index, baseindex) ;
			__m512 at = _mm512_load_ps ((void*)(a + 16 * k)) ;
			__m512 bt = _mm512_load_ps ((void*)(b + 16 * k)) ;
			accum = _mm512_add_ps (accum, at) ;
			accum = _mm512_add_ps (accum, bt) ;
		}
		index = _mm512_set1_epi32(16 * omp_get_thread_num());
		_mm512_i32scatter_ps (c, index, accum, sizeof (float)) ;
	}
	#else
	#endif
}

struct timer
{
	struct timeval _start ;
	struct timeval _stop ;

	void start ()
	{
		::gettimeofday (&_start, 0) ;
	}
	void stop ()
	{
		::gettimeofday (&_stop, 0) ;
	}
	double duration ()
	{
		return (((double)_stop.tv_usec - (double)_start.tv_usec) * 1.0e-6) +
		(double)_stop.tv_sec - (double)_start.tv_sec ;
	}
} ;

void* memalign(int align, int size)
{
	char* tmp = (char*) ::malloc (size + align) ;
	size_t itmp = (size_t) tmp ;
	itmp = (itmp + align -1) & (~(align-1));
	return (void*) itmp ;
}

#define ALLOC   alloc_if(1)
#define FREE    free_if(1)
#define RETAIN  free_if(0)
#define REUSE   alloc_if(0)

int main(int argc, char* argv[])
{
	timer naive, gather, load, outer ;
	int redocount = 100 ;
	int size = 128*1024*1024 ;
	// int size = 8*1024*1024 ;
	float* a = (float*)memalign(64, size * sizeof (float)) ;
	float* b = (float*)memalign(64, size * sizeof (float)) ;
	float* cnaive = (float*)memalign(64, size * sizeof (float)) ;
	float* cgather = (float*)memalign(64, size * sizeof (float)) ;
	float* cload = (float*)memalign(64, size * sizeof (float)) ;
	for (int k = 0 ; k < size ; ++k)
	{
		a [k] = (float)k ;
		b [k] = 2.0f * (float)k ;
		cnaive [k] = 0.0f ;
		cgather [k] = 0.0f ;
		cload [k] = 0.0f ;
	}

	outer.start();
	
	// copy/alloc to device
	#pragma offload target(mic:0) in(a:length(size) ALLOC RETAIN)
	{}
	#pragma offload target(mic:0) in(b:length(size) ALLOC RETAIN)
	{}
	#pragma offload target(mic:0) out(cnaive:length(size) ALLOC RETAIN)
	{}
	#pragma offload target(mic:0) out(cgather:length(size) ALLOC RETAIN)
	{}
	#pragma offload target(mic:0) out(cload:length(size) ALLOC RETAIN)
	{}

	::fprintf (stderr, "[INFO] : blank run\n");
	#pragma offload target(mic:0) nocopy(a:length(size) REUSE RETAIN) nocopy(b:length(size) REUSE RETAIN) nocopy(cnaive:length(size) REUSE RETAIN) nocopy(cgather:length(size) REUSE RETAIN) nocopy(cload:length(size) REUSE RETAIN)
	{
		addition_naive (size, a, cnaive) ;
		addition_gather (size, a, b, cgather) ;
		addition_load (size, a, b, cload) ;
	}
	::fprintf (stderr, "[INFO] : starting inner\n");

	naive.start();

	for (int k = 0 ; k < redocount ; ++k)
	{
		#pragma offload target(mic:0) nocopy(a:length(size) REUSE RETAIN) nocopy(cnaive:length(size) REUSE RETAIN)
		addition_naive (size, a, cnaive) ;
	}

	naive.stop();

	gather.start();

	for (int k = 0 ; k < redocount ; ++k)
	{
		#pragma offload target(mic:0) nocopy(a:length(size) REUSE RETAIN) nocopy(b:length(size) REUSE RETAIN) nocopy(cgather:length(size) REUSE RETAIN)
		addition_gather (size, a, b, cgather) ;
	}

	gather.stop();

	load.start();

	for (int k = 0 ; k < redocount ; ++k)
	{
		#pragma offload target(mic:0) nocopy(a:length(size) REUSE RETAIN) nocopy(b:length(size) REUSE RETAIN) nocopy(cload:length(size) REUSE RETAIN)
		addition_load (size, a, b, cload) ;
	}

	load.stop();

	::fprintf (stderr, "[INFO] : done inner\n");

	// copyback and free
	#pragma offload target(mic:0) nocopy(a:length(size) REUSE FREE) nocopy(b:length(size) REUSE FREE) out(cnaive:length(size) REUSE FREE) out(cgather:length(size) REUSE FREE) out(cload:length(size) REUSE FREE)
	{}

	outer.stop();

	/*
	for (int k = 0 ; k < size ; ++k)
	{
		if (c [k] != (a [k] + b[k]))
		{
			::fprintf (stderr, "Failure at index %d %f <> %f\n", k, c[k], a[k]+b[k]);
			return 1 ;
		}
	}
	::fprintf (stderr, "Success -- took %lf seconds [%lf]\n", inner.duration(), outer.duration()) ;*/
	double tnaive = naive.duration() ;
	double tgather = gather.duration() ;
	double tload = load.duration() ;
	double totalbytes = size * sizeof(float) * redocount ;
	double BWr = 2.0 * totalbytes / (1000.0*1000.0*1000.0) ;
	::fprintf (stderr, "NAIVE Bandwidth = %lf GB/s R\n", BWr / 2.0 / tnaive) ;
	::fprintf (stderr, "GATHER Bandwidth = %lf GB/s R\n", BWr / tgather) ;
	::fprintf (stderr, "LOAD Bandwidth = %lf GB/s R\n", BWr / tload) ;
	return 0 ;
}

