/* (c)2013 ALTIMESH & HighPerformanceComputing Magazine France - All rights reserved */

#include <cstdio>
#include <cstdlib>
#include <immintrin.h>
#include <sys/time.h>

#include <omp.h>

__attribute__((target(mic))) void kernel_naive(int size, const float* __restrict a, int cache_shift, int numidx, const int* __restrict index, float* __restrict b)
{
	#ifdef __MIC__
	#pragma omp parallel for schedule(static,16*(((size+15)/16)/64/omp_get_num_threads()))
	#pragma simd
	#pragma vector aligned
	for (int k = 0 ; k < size ; ++k)
	{
		float tmp = 0.0f ;
		for (int j = 0 ; j < numidx ; ++j)
		{
			tmp += a [cache_shift + k + index [j]] ;
		}
		b [k] = tmp ;
	}
	#else
	#endif
}

__attribute__((target(mic))) void kernel_load(int size, const float* __restrict a, int cache_shift, int numidx, const int* __restrict index, float* __restrict b)
{
	#ifdef __MIC__
	#pragma omp parallel for
	for (int k = 0 ; k < (size/16) ; ++k)
	{
		__m512 tmp = _mm512_set1_ps (0.0f) ;
		for (int j = 0 ; j < numidx ; ++j)
		{
			tmp = _mm512_add_ps (tmp, _mm512_load_ps ((void*)(a + cache_shift + 16*k + index [j]))) ;
		}
		_mm512_store_ps ((void*)(b + 16*k), tmp) ;
	}
	#else
	#endif
}

#include <omp.h>

struct timer
{
	struct timeval _start ;
	struct timeval _stop ;

	void start ()
	{
		::gettimeofday (&_start, 0) ;
	}
	void stop ()
	{
		::gettimeofday (&_stop, 0) ;
	}
	double duration ()
	{
		return (((double)_stop.tv_usec - (double)_start.tv_usec) * 1.0e-6) +
		(double)_stop.tv_sec - (double)_start.tv_sec ;
	}
} ;

void* memalign(int align, int size)
{
	char* tmp = (char*) ::malloc (size + align) ;
	size_t itmp = (size_t) tmp ;
	itmp = (itmp + align -1) & (~(align-1));
	return (void*) itmp ;
}

#define ALLOC   alloc_if(1)
#define FREE    free_if(1)
#define RETAIN  free_if(0)
#define REUSE   alloc_if(0)

int main(int argc, char* argv[])
{
	int itemsize = 1024 ;
	if (argc > 1)
		itemsize = atoi (argv[1]) ;
	timer naive, load, outer ;	
	int redocount = 20 ;
	int size = 240*itemsize ;
	int numidx = 60 ;
	// int size = 8*1024*1024 ;
	float* a = (float*)memalign(64, size * numidx * redocount * sizeof (float)) ;
	float* bnaive = (float*)memalign(64, size * sizeof (float)) ;
	float* bload = (float*)memalign(64, size * sizeof (float)) ;
	int* index = (int*)memalign(64, numidx * sizeof (float)) ;
	for (int i = 0 ; i < size ; ++i)
	{
		for (int k = 0 ; k < numidx ; ++k)
		{
			a [k * size + i] = 1.0f / (0.5f + (float)(i+k)) ;	
			if (i == 0)
				index [k] = ((k*k) % numidx) * size ;
		}
		bnaive [i] = 0.0f ;
		bload [i] = 0.0f ;
	}

	outer.start();
	
	// copy/alloc to device
	#pragma offload target(mic:0) in(a:length(size*redocount*numidx) ALLOC RETAIN)
	{}
	#pragma offload target(mic:0) out(bnaive:length(size) ALLOC RETAIN)
	{}
	#pragma offload target(mic:0) out(bload:length(size) ALLOC RETAIN)
	{}
	#pragma offload target(mic:0) in(index:length(numidx) ALLOC RETAIN)
	{}

	//::fprintf (stderr, "[INFO] : blank run\n");
	#pragma offload target(mic:0) nocopy(a:length(size*redocount*numidx) REUSE RETAIN) nocopy(bnaive:length(size) REUSE RETAIN) nocopy(bload:length(size) REUSE RETAIN) nocopy(index:length(numidx) REUSE RETAIN)
	{
		kernel_naive(size, a, ((redocount-1)*numidx*size), numidx, index, bnaive) ;
		kernel_load(size, a, ((redocount-1)*numidx*size), numidx, index, bload) ;
	}
	//::fprintf (stderr, "[INFO] : starting inner\n");


	double tnaive = 1e12 ;
	for (int k = 0 ; k < redocount ; ++k)
	{
		naive.start();
		#pragma offload target(mic:0) nocopy(a:length(size*redocount*numidx) REUSE RETAIN) nocopy(bnaive:length(size) REUSE RETAIN) nocopy(index:length(numidx) REUSE RETAIN)
		kernel_naive(size, a, k*numidx*size, numidx, index, bnaive) ;
		naive.stop ();
		if (naive.duration () < tnaive) tnaive = naive.duration () ;
	}

	double tload = 1e12 ;
	for (int k = 0 ; k < redocount ; ++k)
	{
		load.start();
		#pragma offload target(mic:0) nocopy(a:length(size*redocount*numidx) REUSE RETAIN) nocopy(bload:length(size) REUSE RETAIN) nocopy(index:length(numidx) REUSE RETAIN)
		kernel_load(size, a, k*numidx*size, numidx, index, bload) ;
		load.stop();
		if (load.duration () < tload) tload = load.duration () ;
	}


	//::fprintf (stderr, "[INFO] : done inner\n");

	// copyback and free
	#pragma offload target(mic:0) nocopy(a:length(size*redocount*numidx) REUSE FREE) out(bnaive:length(size) REUSE FREE) out(bload:length(size) REUSE FREE) nocopy(index:length(numidx) REUSE FREE)
	{}

	outer.stop();

	double totalbytes = size * numidx * sizeof(float) ;
	double BWr = totalbytes / (1000.0*1000.0*1000.0) ;
	::fprintf (stderr, "[%d] = NAIVE Bandwidth = %lf GB/s R [%lf s]\n", itemsize, BWr / tnaive, tnaive) ;
	::fprintf (stderr, "[%d] = LOAD Bandwidth = %lf GB/s R [%lf s]\n", itemsize, BWr / tload, tload) ;
	return 0 ;
}

